import 'dart:typed_data';

import 'package:encrypt/encrypt.dart';
import 'package:pointycastle/digests/sha512.dart';
import 'package:pointycastle/ecc/api.dart';

import 'IkeyPair.dart';
import 'Seed.dart';
import '../utils/Util.dart';
import 'dart:convert' as convert;

class Ecies{
  static const int AESKeyLength = 32;
  static const int HMACKeyLength = 32;
  static const int IVLength = 16; // bytes
  static const int PUBLIC_LENGTH = 33;
  static final int AESBlockLength = 16;
  static final int MACBYTELENGTH = 32;

  static Uint8List eciesEncrypt(Uint8List plainBytes,Uint8List pub){
    var seed = Seed.randomSeed();
    var ephKeyPair = seed.keyPair();

    var secret = doECDH(ephKeyPair.getPriv(), pub);
    SHA512Digest digest = new SHA512Digest();
    var kdfOutput = digest.process(secret);


    var hmacKey = kdfOutput.sublist(AESKeyLength, AESKeyLength + HMACKeyLength);
    var macBytes = Util.hmac(hmacKey, plainBytes);

    Uint8List plainBuf = new Uint8List(macBytes.length + plainBytes.length);
    Util.arraycopy(macBytes, 0, plainBuf, 0, macBytes.length);
    Util.arraycopy(plainBytes, 0, plainBuf, macBytes.length, plainBytes.length);

    var aesKey = kdfOutput.sublist(0,AESKeyLength);
    final key = new Key(aesKey);
    final iv = new IV(Util.getRandomBytes(IVLength));
    final encrypter = Encrypter(AES(key, mode: AESMode.cbc));
    final encrypted = encrypter.encryptBytes(plainBuf, iv: iv);
    var encryptedBytes = encrypted.bytes;

    Uint8List finalBytes = new Uint8List(PUBLIC_LENGTH + IVLength + encryptedBytes.length);
    Util.arraycopy(ephKeyPair.canonicalPubBytes(),0,finalBytes,0,PUBLIC_LENGTH);
    Util.arraycopy(iv.bytes,0,finalBytes,PUBLIC_LENGTH,IVLength);
    Util.arraycopy(encryptedBytes,0,finalBytes,PUBLIC_LENGTH + IVLength,encryptedBytes.length);

    return finalBytes;
  }

  static Uint8List eciesDecrypt(Uint8List ciphertext,Seed seed){
    int level = 256;
    int D_len = level >> 3;
    int ct_len = ciphertext.length;
    int totalLen = PUBLIC_LENGTH + IVLength + D_len + AESBlockLength;
    if (ct_len < totalLen){
      throw new Exception("Illegal cipherText length: " + ct_len.toString() + " must be >= " + totalLen.toString());
    }
    Uint8List publicOther = new Uint8List(PUBLIC_LENGTH);
    Uint8List iv = new Uint8List(IVLength);
    Uint8List cipherBytes = new Uint8List(ciphertext.length - PUBLIC_LENGTH - IVLength);
    Util.arraycopy(ciphertext, 0, publicOther, 0, PUBLIC_LENGTH);
    Util.arraycopy(ciphertext, PUBLIC_LENGTH, iv, 0, IVLength);
    Util.arraycopy(ciphertext, PUBLIC_LENGTH + IVLength, cipherBytes, 0, ciphertext.length - PUBLIC_LENGTH - IVLength);

    IKeyPair pair = seed.keyPair();

    var secret = doECDH(pair.getPriv(), publicOther);
    SHA512Digest digest = new SHA512Digest();
    var kdfOutput = digest.process(secret);

    Uint8List aesKey = kdfOutput.sublist(0,AESKeyLength);
    Uint8List hmacKey = kdfOutput.sublist(AESKeyLength,AESKeyLength + HMACKeyLength);

    final key = new Key(aesKey);
    final iv2 = new IV(iv);
    final encrypter = Encrypter(AES(key, mode: AESMode.cbc));
    Encrypted encrypted = new Encrypted(cipherBytes);
    Uint8List decryptedBytes = Uint8List.fromList(encrypter.decryptBytes(encrypted, iv: iv2));
    Uint8List macBytes = new Uint8List(MACBYTELENGTH);
    Uint8List plainText = new Uint8List(decryptedBytes.length - MACBYTELENGTH);
    Util.arraycopy(decryptedBytes, 0, macBytes, 0, MACBYTELENGTH);
    Util.arraycopy(decryptedBytes, macBytes.length, plainText, 0, plainText.length);

    Uint8List recoveredD = Util.hmac(hmacKey, plainText);
    if(recoveredD.toString() != macBytes.toString()){
      return null;
    }

    return plainText;
  }

  static Uint8List doECDH(BigInt priv,Uint8List pub){
    var eccDomain = new ECDomainParameters("secp256k1");
    ECCurve curve = eccDomain.curve;
    ECPoint pubPoint = curve.decodePoint(pub);
    ECPublicKey pubKey = new ECPublicKey(pubPoint,eccDomain);
    var point = pubKey.Q * priv;
    return Util.createUint8ListFromBigInt(point.x.toBigInteger());
  }
}